import getpass

import click

import komodo_cli.printing as printing
from komodo_cli.utils import APIClient, handle_errors

MIN_PASSWORD_LENGTH = 8


@click.command("register")
@handle_errors
def cmd_register():
    email = input("Email: ")

    while True:
        pw1 = getpass.getpass("Password: ")
        if len(pw1) < MIN_PASSWORD_LENGTH:
            printing.error(
                "Password must be at least 8 characters, please enter a new password."
            )
            continue
        pw2 = getpass.getpass("Re-enter password: ")

        if pw1 != pw2:
            printing.error("Passwords do not match, please try again.")
            continue

        break

    APIClient.register(email, pw1)

    printing.success(f"Successfully registered user {email}")



import os

import click
import yaml

import komodo_cli.printing as printing
from komodo_cli.commands.cmd_run import _run_helper
from komodo_cli.utils import handle_errors


@click.command("do")
@click.option("--file", "-f", type=str, required=True)
@click.option("--detach", "-d", is_flag=True)
@click.pass_context
@handle_errors
def cmd_do(ctx: click.Context, file: str, detach: bool):
    if not os.path.isfile(file):
        printing.error(
            f"{file} does not exist",
            bold=True,
        )

    with open(file, "r") as f:
        action = yaml.load(f, yaml.FullLoader)

    kind = action["kind"]

    if kind == "task":
        # TODO: use vyper for defaults
        backend = action.get("backend", None)
        resource = action.get("resource", None)
        num_nodes = action.get("num_nodes", 1)
        backend_config_override = action.get("backend_config", {})
        command = action["command"]
        _run_helper(
            ctx,
            backend,
            resource,
            num_nodes,
            backend_config_override,
            detach,
            command,
        )
    # TODO: add support for machine actions
    else:
        printing.error(f"Action of type {kind} is not supported", bold=True)
        exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This is the entry point for the command-line interface (CLI) application.

It can be used as a handy facility for running the task from a command line.

.. note::

    To learn more about Click visit the
    `project website <http://click.pocoo.org/5/>`_.  There is also a very
    helpful `tutorial video <https://www.youtube.com/watch?v=kNke39OZ2k0>`_.

    To learn more about running Luigi, visit the Luigi project's
    `Read-The-Docs <http://luigi.readthedocs.io/en/stable/>`_ page.

.. currentmodule:: komodo_cli.cli
.. moduleauthor:: kmushegi <mushegiani@gmail.com>
"""
import os
import sys

import click
from loguru import logger
from vyper import Vyper

import komodo_cli.printing as printing
from komodo_cli.commands.api_key.api_key import api_key
from komodo_cli.commands.backend.backend import backend
from komodo_cli.commands.cmd_cancel import cmd_cancel
from komodo_cli.commands.cmd_delete import cmd_delete
from komodo_cli.commands.cmd_describe import cmd_describe
from komodo_cli.commands.cmd_do import cmd_do
from komodo_cli.commands.cmd_list import cmd_list
from komodo_cli.commands.cmd_logs import cmd_logs
from komodo_cli.commands.cmd_register import cmd_register
from komodo_cli.commands.cmd_run import cmd_run
from komodo_cli.commands.cmd_shell import cmd_shell
from komodo_cli.commands.compute.compute import compute
from komodo_cli.commands.init.init import init
from komodo_cli.commands.machine.machine import machine
from komodo_cli.overlay_images.overlay_images import OverlayImages
from komodo_cli.utils import handle_errors

from .__init__ import __version__

LOGGING_LEVELS = {
    0: "NOTSET",
    1: "ERROR",
    2: "WARNING",
    3: "INFO",
    4: "DEBUG",
}  #: a mapping of `verbose` option counts to logging levels


@click.group()
@click.option("--verbose", "-v", count=True, help="Enable verbose output.")
@click.pass_context
@handle_errors
def cli(ctx: click.Context, verbose: int) -> None:
    """Run Komodo."""

    # Use the verbosity count to determine the logging level...
    if verbose > 0:
        logger.info(f"verbose count {verbose}")
        log_level = LOGGING_LEVELS[verbose] if verbose in LOGGING_LEVELS else "DEBUG"
        logger.remove()
        logger.add(
            sys.stdout,
            level=log_level,
        )
        printing.warning(
            f"Verbose logging is enabled. " f"(LEVEL={log_level})",
        )
    else:
        logger.remove()

    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose


@cli.command()
def version():
    """Get the library version."""
    printing.info(f"{__version__}", bold=True)


cli.add_command(cmd_register)
cli.add_command(init)
cli.add_command(cmd_run)
cli.add_command(cmd_list)
cli.add_command(cmd_logs)
cli.add_command(cmd_describe)
cli.add_command(cmd_shell)
cli.add_command(cmd_cancel)
cli.add_command(cmd_delete)
cli.add_command(machine)
cli.add_command(cmd_do)
cli.add_command(compute)
cli.add_command(api_key)
cli.add_command(backend)



import click
from loguru import logger

import komodo_cli.printing as printing
from komodo_cli.backends.backend import Backend
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.types import (
    JobNotFoundException,
    JobNotInDesiredStateException,
    JobStatus,
)
from komodo_cli.utils import APIClient, handle_errors, update_context_with_api_client


@click.command("delete")
@click.argument(
    "job_id",
    type=str,
)
@click.option(
    "--force",
    "-f",
    type=bool,
    is_flag=True,
    help="Set this flag if you want to delete the job even if the job is still running",
)
@click.pass_context
@handle_errors
def cmd_delete(ctx: click.Context, job_id: str, force: bool):
    """Delete a Komodo job."""
    update_context_with_api_client(ctx)
    logger.info(f"Deleting job {job_id}")

    api_client: APIClient = ctx.obj["api_client"]

    if job_id == "all":
        jobs = api_client.list_jobs(limit=-1)
    else:
        jobs = [api_client.get_job(job_id)]

    for job in jobs:
        backend_schema = api_client.get_backend(job.backend_name)
        backend: Backend = BackendFactory.get_backend(
            backend_schema,
            api_client,
        )

        printing.warning(f"Deleting job {job.id}", bold=True)
        api_client.delete_job(job.id, force)
        # only the local backend actually does anything here
        backend.delete(job.backend_job_id)



import json
import os
import sys
from typing import Dict, List, Union
from uuid import getnode

import click
import docker
import git
import requests
import yaml
from loguru import logger
from vyper import Vyper

import komodo_cli.printing as printing
from komodo_cli.types import (
    Backend,
    ClientException,
    Job,
    JobNotFoundException,
    JobStatus,
    Machine,
    Resource,
)

API_BASE_URL = os.environ.get(
    "KOMODO_API_URL", "https://komodo-api-wprtgx67oa-uc.a.run.app"
)


class NoCredentialsException(Exception):
    pass


LOCAL_STATUS_MAPPING = {
    "cancelling": JobStatus.CANCELLING,
    "cancelled": JobStatus.CANCELLED,
    "created": JobStatus.PENDING,
    "running": JobStatus.RUNNING,
    "paused": JobStatus.RUNNING,
    "restarting": JobStatus.PENDING,
    "removing": JobStatus.FINISHED,
    "exited": JobStatus.FINISHED,
    "dead": JobStatus.FINISHED,
}


class APIClient:
    # TODO: Support API key from a KOMODO_API_KEY env variable
    def __init__(
        self,
        api_key=None,
        username=None,
        password=None,
    ):
        self.api_base_url = API_BASE_URL
        self.username = username
        self.password = password

        if api_key is None:
            if username is None or password is None:
                raise NoCredentialsException(
                    "API key or username & password are required for authentication."
                )
            else:
                try:
                    self.access_token, self.token_type = self.create_token()
                except Exception as e:
                    raise e
        else:
            self.access_token = api_key
            self.token_type = "Bearer"

    @classmethod
    def api_request(
        cls,
        method: str,
        url: str,
        headers: Dict = None,
        files: Dict = None,
        data: Dict = None,
    ) -> Union[Dict, List]:  # Not using | for > 1 return type for < Py 3.10 compat
        # Generic way to make HTTP requests w/ standard error handling
        # TODO: debug-level info logs?
        response = requests.request(
            method,
            url,
            headers=headers,
            files=files,
            data=data,
        )
        response.raise_for_status()

        retval = response.json()
        return retval

    @classmethod
    def register(cls, email: str, password: str):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        payload = json.dumps(
            {
                "email": email,
                "password": password,
            },
        )

        cls.api_request(
            "POST",
            f"{API_BASE_URL}/api/v1/auth/register",
            headers=headers,
            data=payload,
        )

    def create_token(self):
        headers = {
            # requests won't add a boundary if this header is set when you pass files=
            # 'Content-Type': 'multipart/form-data',
        }

        files = {
            "username": (None, self.username),
            "password": (None, self.password),
        }

        auth = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/auth/token/login",
            headers=headers,
            files=files,
        )

        return auth["access_token"], auth["token_type"]

    def get_all_tokens(self):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        tokens = self.api_request(
            "GET",
            f"{self.api_base_url}/api/v1/auth/token/list",
            headers=headers,
        )

        return tokens

    def delete_token(self, token: str):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        self.api_request(
            "DELETE",
            f"{self.api_base_url}/api/v1/auth/token/{token}",
            headers=headers,
        )

    def create_backend(self, name: str, type: str, config: dict) -> Backend:
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        payload = json.dumps(
            {
                "name": name,
                "type": type,
                "config": config,
            },
        )

        backend_json = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/backends",
            headers=headers,
            data=payload,
        )

        backend = Backend.from_dict(backend_json)
        return backend

    def list_backends(self) -> List[Backend]:
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        backend_jsons = self.api_request(
            "GET",
            f"{self.api_base_url}/api/v1/backends",
            headers=headers,
        )

        backends = [Backend.from_dict(b) for b in backend_jsons]
        return backends

    def get_backend(self, backend_name: str) -> Backend:
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        backend_json = self.api_request(
            "GET",
            f"{self.api_base_url}/api/v1/backends/{backend_name}",
            headers=headers,
        )

        backend = Backend.from_dict(backend_json)
        return backend

    def delete_backend(self, name: str, force: bool):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        self.api_request(
            "DELETE",
            f"{self.api_base_url}/api/v1/backends/{name}?force={str(force).lower()}",
            headers=headers,
        )

    def create_backend_resource(
        self, backend_name: str, resource_name: str, resource_config: dict
    ) -> Resource:
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        payload = json.dumps(
            {
                "name": resource_name,
                "config": resource_config,
            },
        )

        resource_json = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/backends/{backend_name}/resources",
            headers=headers,
            data=payload,
        )

        resource = Resource.from_dict(resource_json)
        return resource

    def delete_backend_resource(self, backend_name: str, resource_name: str):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        self.api_request(
            "DELETE",
            f"{self.api_base_url}/api/v1/backends/{backend_name}/resources/{resource_name}",
            headers=headers,
        )

    def create_job(
        self,
        backend_name: str,
        command: List[str],
        num_nodes: int,
        image: str,
        env: Dict,
        mounts: List,
        workdir: str,
        resource_name: str,
    ) -> Job:
        payload = json.dumps(
            {
                "backend_name": backend_name,
                "command": command,
                "num_nodes": num_nodes,
                "image": image,
                "env": env,
                "mounts": mounts,
                "workdir": workdir,
                "resource_name": resource_name,
            }
        )

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        job_json = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/jobs",
            headers=headers,
            data=payload,
        )
        job_json["status"] = JobStatus(job_json["status"])
        job = Job.from_dict(job_json)

        return job

    def create_machine(
        self,
        machine_name: str,
        backend_name: str,
        image: str,
        env: Dict,
        mounts: List,
        workdir: str,
        resource_name: str,
    ) -> Job:
        payload = json.dumps(
            {
                "name": machine_name,
                "backend_name": backend_name,
                "image": image,
                "env": env,
                "mounts": mounts,
                "workdir": workdir,
                "resource_name": resource_name,
            }
        )

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        machine_json = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/machines",
            headers=headers,
            data=payload,
        )

        machine_json["status"] = JobStatus(machine_json["status"])
        machine = Machine.from_dict(machine_json)

        return machine

    def update_job(
        self,
        job_id: str,
        status: JobStatus,
        backend_job_id: str = None,
    ):
        if status == JobStatus.NOT_AVAILABLE_CURRENT_COMPUTER:
            # this is just a client-side status
            status = None
        payload = json.dumps(
            {
                "status": (status.value if status else None),
                "backend_job_id": backend_job_id,
            }
        )
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        response = self.api_request(
            "PUT",
            f"{self.api_base_url}/api/v1/jobs/{job_id}",
            headers=headers,
            data=payload,
        )

        return response

    def update_machine(
        self,
        machine_name: str,
        status: JobStatus,
        backend_job_id: str = None,
    ):
        if status == JobStatus.NOT_AVAILABLE_CURRENT_COMPUTER:
            # this is just a client-side status
            status = None
        payload = json.dumps(
            {
                "status": (status.value if status else None),
                "backend_job_id": backend_job_id,
            }
        )
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        response = self.api_request(
            "PUT",
            f"{self.api_base_url}/api/v1/machines/{machine_name}",
            headers=headers,
            data=payload,
        )

        return response

    def delete_job(self, job_id: str, force: bool = False):
        job = self.get_job(job_id)
        if job.status == JobStatus.NOT_AVAILABLE_CURRENT_COMPUTER:
            raise ClientException(
                f"Cannot delete local job {job_id} because it was not created on this computer"
            )

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        response = self.api_request(
            "DELETE",
            f"{self.api_base_url}/api/v1/jobs/{job_id}?force={str(force).lower()}",
            headers=headers,
        )

        return response

    def cancel_job(self, job_id: str):
        job = self.get_job(job_id)
        if job.status == JobStatus.NOT_AVAILABLE_CURRENT_COMPUTER:
            raise ClientException(
                f"Cannot delete local job {job_id} because it was not created on this computer"
            )

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        response = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/jobs/cancel/{job_id}",
            headers=headers,
        )

        return response

    def delete_machine(self, machine_name: str, force: bool = False):
        machine = self.get_machine(machine_name)
        if machine.status == JobStatus.NOT_AVAILABLE_CURRENT_COMPUTER:
            raise ClientException(
                f"Cannot delete local machine {machine_name} because it was not created on this computer"
            )

        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        response = self.api_request(
            "DELETE",
            f"{self.api_base_url}/api/v1/machines/{machine_name}?force={force}",
            headers=headers,
        )

        return response

    def _update_local_job_status(self, client, job):
        backend_job_id = job.backend_job_id
        if ":" in backend_job_id:
            mac_address, docker_container_id = backend_job_id.split(":")
        else:
            # backwards compatibility
            docker_container_id = backend_job_id
            mac_address = None

        # TODO: error handling
        if mac_address and str(getnode()) != mac_address:
            status = JobStatus.NOT_AVAILABLE_CURRENT_COMPUTER
        else:
            try:
                container = client.containers.get(docker_container_id)
                status = LOCAL_STATUS_MAPPING[container.status]
            except docker.errors.NotFound:
                status = JobStatus.NOT_FOUND
            except docker.errors.NullResource:
                status = JobStatus.ERROR

        # TODO: make this async
        self.update_job(
            job.id,
            status,
        )
        job.status = status

    def _update_local_machine_status(self, client, machine):
        backend_job_id = machine.backend_job_id

        if ":" in backend_job_id:
            mac_address, docker_container_id = backend_job_id.split(":")
        else:
            # backwards compatibility
            docker_container_id = backend_job_id
            mac_address = None

        if mac_address and str(getnode()) != mac_address:
            status = JobStatus.NOT_AVAILABLE_CURRENT_COMPUTER
        else:
            try:
                container = client.containers.get(docker_container_id)
                status = LOCAL_STATUS_MAPPING[container.status]
            except docker.errors.NotFound:
                status = JobStatus.NOT_FOUND
            except docker.errors.NullResource:
                status = JobStatus.ERROR
        # TODO: make this async
        self.update_machine(
            machine.name,
            status,
        )
        machine.status = status

    def list_jobs(self, skip: int = 0, limit: int = 10):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        response = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/jobs/all?skip={skip}&limit={limit}",
            headers=headers,
        )

        backends = self.list_backends()
        backends = {b.name: b for b in backends}

        jobs: List[Job] = []
        client = None
        for job_json in response:
            job = Job.from_dict(job_json)

            backend_name = job.backend_name
            backend_type = backends[backend_name].type
            if backend_type == "local":
                if client is None:
                    client = docker.from_env()
                self._update_local_job_status(client, job)

            jobs.append(job)

        return jobs

    def list_machines(self):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        response = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/machines/all",
            headers=headers,
        )

        backends = self.list_backends()
        backends = {b.name: b for b in backends}

        machines: List[Machine] = []
        client = None
        for machine_json in response:
            machine = Machine.from_dict(machine_json)

            backend_name = machine.backend_name
            backend_type = backends[backend_name].type
            if backend_type == "local":
                if client is None:
                    client = docker.from_env()
                self._update_local_machine_status(client, machine)
            machines.append(machine)
        return machines

    def get_job(self, job_id: str) -> Job:
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        job_json = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/jobs/{job_id}",
            headers=headers,
        )
        job_json["status"] = JobStatus(job_json["status"])
        job = Job.from_dict(job_json)

        backend = self.get_backend(job.backend_name)

        client = None
        backend_type = backend.type
        if backend_type == "local":
            if client is None:
                client = docker.from_env()
            self._update_local_job_status(client, job)

        return job

    def get_machine(self, machine_name: str) -> Machine:
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"{self.token_type} {self.access_token}",
        }

        machine_json = self.api_request(
            "POST",
            f"{self.api_base_url}/api/v1/machines/{machine_name}",
            headers=headers,
        )
        machine_json["status"] = JobStatus(machine_json["status"])
        machine = Machine.from_dict(machine_json)

        backend = self.get_backend(machine.backend_name)

        client = None
        backend_type = backend.type
        if backend_type == "local":
            if client is None:
                client = docker.from_env()
            self._update_local_machine_status(client, machine)

        return machine


def get_komodo_project_config_path() -> str:
    try:
        project_config_path = os.environ.get(
            "KOMODO_PROJECT_DIR",
            None,
        )
        if not project_config_path:
            # Project config must be at git repo root if not specified via env
            git_repo = git.Repo(os.getcwd(), search_parent_directories=True)
            project_config_path = git_repo.git.rev_parse("--show-toplevel")
        return project_config_path
    except FileNotFoundError:
        printing.error(
            f"Project config not found.\n* Have you run 'komo init'?",
            bold=True,
        )
        sys.exit(1)


def update_context_with_project_config(ctx: click.Context):
    project_config_path = get_komodo_project_config_path()
    v_project_config = Vyper()
    v_project_config.add_config_path(os.path.join(project_config_path, ".komo"))
    v_project_config.set_config_name("project")
    v_project_config.set_config_type("yaml")
    try:
        v_project_config.read_in_config()
    except FileNotFoundError as e:
        logger.error(e)
        printing.error("Project config not found.", bold=True)
        sys.exit(1)

    ctx.obj["project_config"] = v_project_config


def update_context_with_api_client(ctx: click.Context):
    try:
        if os.environ.get("KOMODO_GLOBAL_CONFIG_DIR", None):
            api_key_file = os.path.join(
                os.environ.get("KOMODO_GLOBAL_CONFIG_DIR"), ".komo", "api-key"
            )
        else:
            api_key_file = os.path.join(os.path.expanduser("~"), ".komo", "api-key")
        if not os.path.isfile(api_key_file):
            printing.error(
                f"No api key found at {api_key_file}. If you don't have an api key, run `komodo api-key create` to generate one, and then store it in ~/.komo/api-key."
            )
            exit(1)
        with open(api_key_file, "r") as f:
            api_key = f.read().strip()
        c = APIClient(api_key=api_key)
        ctx.obj["api_client"] = c
    except Exception as e:
        logger.error(e)
        printing.error("Error initializing API client", bold=True)
        sys.exit(1)


def config_args_to_dict(config_args: str):
    config = {}

    for arg in config_args.split(","):
        if not arg:
            continue

        if not "=" in arg or arg.count("=") > 1:
            printing.error(
                f"Invalid config parameter {arg}. Should be in form param=value.",
                bold=True,
            )
            sys.exit(1)

        param, value = arg.split("=")
        value = yaml.safe_load(value)
        config[param] = value

    return config


def update_context_project_config_with_backend_override(
    ctx: click.Context, backend_name: str, backend_config_override: dict
):
    backend_config = ctx.obj["project_config"].get("backends").get(backend_name, None)
    if backend_config is None:
        raise ClientException(
            f"Backend {backend_name} doesn't exist in the project config"
        )

    for key, value in backend_config_override.items():
        backend_config[key] = value


def handle_errors(fn):
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ClientException as e:
            printing.error(str(e))
            exit(1)
        except requests.exceptions.HTTPError as e:
            # Catch HTTP errors like 401, 500 etc.
            text = e.response.text
            try:
                j = e.response.json()
            except requests.exceptions.JSONDecodeError:
                j = None
            if j:
                if "detail" in j:
                    text = j["detail"]
            printing.error(f"Got HTTP Error Code {e.response.status_code}: {text}")
            exit(1)
        except requests.exceptions.ConnectionError as ec:
            printing.error(f"Connection Error: {str(ec)}")
            exit(1)
        except requests.exceptions.Timeout as et:
            printing.error(f"Timeout Error: {str(et)}")
            exit(1)
        except requests.exceptions.RequestException as e:
            # Panic
            printing.error(f"Unknown Error: {str(e)}")
            exit(1)
        except Exception as e:
            logger.error(str(e))
            printing.error(f"Encountered an unknown error: {str(e)}")
            raise e

    return inner



import os
import time
from typing import Optional

import click
from loguru import logger
from vyper import Vyper

import komodo_cli.printing as printing
from komodo_cli.backends.backend import Backend
from komodo_cli.backends.backend_factory import BackendFactory
from komodo_cli.overlay_images.overlay_images import OverlayImages
from komodo_cli.types import JobNotFoundException, JobStatus, TooManyNodes
from komodo_cli.utils import (
    APIClient,
    config_args_to_dict,
    handle_errors,
    update_context_project_config_with_backend_override,
    update_context_with_api_client,
    update_context_with_project_config,
)


@click.command("run")
@click.option(
    "--backend",
    "-b",
    type=str,
    default=None,
    help="Name of the backend to use.",
)
@click.option(
    "--resource",
    "-r",
    type=str,
    default=None,
    help="Resource type to use for the job as defined in the config file. Ignored for local jobs.",
)
@click.option(
    "--num_nodes",
    "-n",
    type=int,
    default=1,
    help="Number of nodes for the job",
)
@click.option(
    "--backend-config",
    "-c",
    type=str,
    default="",
    help=(
        "Override any parameters in the 'config' section of the corresponding backend section in your .komo/project.yaml "
        "config file. For example, if you are using your 'local' backend, and want to change the workdir, you would do "
        "--backend-config workdir=/my-workdir. You can override multiple parameters by separating them with commas. For "
        "example: --backend-config workdir=/my-workdir,base_image=python:3.10"
    ),
)
@click.option("--detach", "-d", is_flag=True)
@click.argument("args", nargs=-1)
@click.pass_context
@handle_errors
def cmd_run(
    ctx: click.Context,
    backend: Optional[str],
    resource: Optional[str],
    num_nodes: int,
    backend_config: str,
    detach: bool,
    args,
):
    """Run a Komodo job."""
    backend_config_override = config_args_to_dict(backend_config)
    _run_helper(
        ctx,
        backend,
        resource,
        num_nodes,
        backend_config_override,
        detach,
        args,
    )


def _run_helper(
    ctx: click.Context,
    backend_name: Optional[str],
    resource_name: Optional[str],
    num_nodes: int,
    backend_config_override: dict,
    detach: bool,
    args,
):
    update_context_with_api_client(ctx)
    update_context_with_project_config(ctx)

    project_config: Vyper = ctx.obj["project_config"]
    if backend_name is None:
        backend_name = project_config.get("default_backend")
        if backend_name is None:
            printing.error("No backend was specified", bold=True)
            exit(1)

    api_client: APIClient = ctx.obj["api_client"]
    backend_schema = api_client.get_backend(backend_name)
    backend: Backend = BackendFactory.get_backend(
        backend_schema,
        api_client,
    )

    update_context_project_config_with_backend_override(
        ctx, backend_name, backend_config_override
    )

    project_backend_config = project_config.get("backends")[backend_name]
    if resource_name is None:
        resource_name = project_backend_config.get("default_resource")
        if resource_name is None:
            # resource not needed for local backend
            if backend_schema.type != "local":
                printing.error("No resource was specified", bold=True)
                exit(1)

    logger.info(f"Using backend {backend_name}")
    logger.info(f"Resource: {resource_name}")
    logger.info(f"Command: ({', '.join(args)})")

    workspace = project_backend_config.get("workspace", os.getcwd())
    if workspace:
        o = OverlayImages(
            project_backend_config["base_image"],
            workspace,
            project_backend_config.get("workdir"),
            project_config.get("overlay_images_repository"),
        )

        printing.info("Building an overlay image", bold=True)
        overlay_image = o.build_overlay_image()

        printing.info(
            f"Pushing an overlay image to repository {o.overlay_images_repository}",
            bold=True,
        )

        o.push_overlay_image(overlay_image)
        project_backend_config["image"] = overlay_image.tags[0]
    else:
        project_backend_config["image"] = project_backend_config["base_image"]

    printing.info("Starting a job", bold=True)
    job = backend.run(
        args,
        num_nodes,
        resource_name,
        project_backend_config["image"],
        project_backend_config["env"],
        project_backend_config["mounts"],
        project_backend_config["workdir"],
    )

    printing.success(
        f"Created a {job.backend_name} job with ID {job.id}",
        bold=True,
    )

    if detach:
        return

    printing.info("Waiting for job to start...", bold=True)
    job_id = job.id

    while True:
        job = api_client.get_job(job_id)
        status = job.status
        if status != JobStatus.PENDING:
            break
        time.sleep(1)

    printing.success("Job started successfully")

    for line in backend.logs(job.backend_job_id, 0, True):
        if type(line) != str:
            line = line.decode("utf-8")
        printing.info(f"{line.strip()}")
